#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <string.h>

#include "basso_commons.h"
#include "BasicMatrixOperations.h"
#include "MemoryManagement.h"
#include "../../include/DataTypes.h"

/**
 * Copies the data from the row-wise saved A->matrix to
 * the col-wise saved A->matrixCol.
 * We assume that memory for A->matrixCol was allocated
 * previously.
 */
basso_bmatrix_cm *restrict
colWiseMatrixFromRowWiseMatrix(const basso_bmatrix_rm *restrict A)
{
	const uint64_t mRows = A->mRows;
	const uint64_t nCols = A->nCols;

	const uint64_t integerPerRow = A->integerPerRow;

	basso_bmatrix_cm* B = checkedCallocBitMatColWise(mRows, nCols);

	const uint64_t integerPerCol = B->integerPerCol;

	// copy all 1-entries from A->matrix to A->matrixCol
	for (uint64_t i=0; i<mRows; i++) {
		for (uint64_t j=0; j<nCols; j++) {
			// if A[i,j] = 1 set it to one
			if ( (A->matrixRow[i*integerPerRow + j/64] & (ONE << j%64)) > 0) {
				B->matrixCol[j*integerPerCol + i/64] |= (ONE << i%64);
			}
		}
	}

	return B;
}

/**
 * Copies the data from the col-wise saved A->matrixCol to
 * the row-wise saved A->matrix.
 * We assume that memory for A->matrix was allocated
 * previously.
 */
basso_bmatrix_rm *restrict
rowWiseMatrixFromColWiseMatrix(const basso_bmatrix_cm *restrict A)
{
	const uint64_t mRows = A->mRows;
	const uint64_t nCols = A->nCols;

	const uint64_t integerPerCol = A->integerPerCol;

	basso_bmatrix_rm* B = checkedCallocBitMatRowWise(mRows, nCols);

	const uint64_t integerPerRow = B->integerPerRow;

	// copy all 1-entries from A->matrixCol to A->matrix
	for (uint64_t j=0; j<nCols; j++) {
		for (uint64_t i=0; i<mRows; i++) {
			// if A[i,j] = 1 set it to 1
			if ( (A->matrixCol[j*integerPerCol + i/64] & (ONE << i%64)) > 0) {
				B->matrixRow[i*integerPerRow + j/64] |= (ONE << j%64);
			}
		}
	}

	return B;
}

/**
 * Returns the number of non-zeros in a row-major matrix
 */
uint64_t
basso_count_nnz_rm(const basso_bmatrix_rm *restrict A)
{
  uint64_t nnz, i, j;
  nnz = 0;
  for (i=0; i < A->mRows; i++) {
    for (j=0; j < A->nCols; j++) {
      if (basso_get_element_rm(A, i, j)) nnz++;
    }
  }
  return nnz;
}

/**
 * Returns the number of non-zeros in a column-major matrix
 */
uint64_t
basso_count_nnz_cm(const basso_bmatrix_cm *restrict A)
{
  uint64_t nnz, i, j;
  nnz = 0;
  for (j=0; j < A->nCols; j++) {
    for (i=0; i < A->mRows; i++) {
      if (basso_get_element_cm(A, i, j)) nnz++;
    }
  }
  return nnz;
}

