#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdbool.h>
#include <string.h> // for memset

#include "basso_commons.h"
#include "BasicMatrixOperations.h"
#include "BinaryHeap.h"
#include "MemoryManagement.h"
#include "../../include/DataTypes.h"
#include "../../include/SelectBasis.h"



/* Public variable that can be used to stop the loop */
volatile sig_atomic_t basso_stop_select_basis = 0;

/**
 * Computes a decomposition of A = BoX, where the k columns of B
 * are picked from the columns of matrix C.
 *
 * Assumes that B and X were not initialized before and stores the 
 * output inside those matrices.
 *
 * Note on the heap usage:
 * The heap does not store the ''real'' score of the column, since
 * this score is not monotonically decreasing (e.g. by overcovering
 * some 0s due to picking other columns, a score might increase).
 * Instead, we store the maximum number of 1s a column may still
 * contribute regardless of overcovering. This number must be
 * monotonically decreasing as we add new columns to the basis.
 */
void
basso_selectBasis(const basso_bmatrix_cm *restrict A,
				  const basso_bmatrix_cm *restrict C,
				  const uint64_t k,
				  const uint64_t weight,
				  basso_bmatrix_cm **restrict Bptr,
				  basso_bmatrix_rm **restrict Xptr)
{
	const uint64_t mRows = A->mRows;
	const uint64_t nCols = A->nCols;

	basso_bmatrix_cm *restrict B = checkedCallocBitMatColWise(A->mRows, k);
	basso_bmatrix_rm *restrict X = checkedCallocBitMatRowWise(k, A->nCols);

	*Bptr = B;
	*Xptr = X;
	
	basso_bmatrix_cm *restrict Covered = checkedCallocBitMatColWise(mRows, nCols);
	uint64_t *restrict bestX = checkedCallocUInt64_t(X->integerPerRow);

	// initialise heap with all columns having max score
	// see general note on heap usage on top of the file
	BinaryHeap *restrict heap = createBinaryHeap(C->nCols);
	for (uint64_t i=0; i<C->nCols; i++) {
		insert(heap, UINT64_MAX, i);
	}
	BinaryHeap *restrict updatedHeap = createBinaryHeap(C->nCols);

	for (uint64_t l=0; l<k && !basso_stop_select_basis; ++l) {
		uint64_t bestCol = UINT64_MAX;
		uint64_t bestScore = 0;
		memset(bestX, 0, sizeof(uint64_t) * X->integerPerRow);

		bool abort = false;

		#pragma omp parallel for 
		for (uint64_t colOfC=0; colOfC < C->nCols; ++colOfC) {
			#pragma omp flush (abort)
			if (! abort) {
				uint64_t currentColOfC = 0;

				#pragma omp critical(getCandidateColumn)
				{
					if (heap->numberOfElements == 0) {
						abort = true;
						#pragma omp flush (abort)
					} else {
						HeapElement element = deleteMax(heap);

						if (element.score < bestScore) {
							// insert element again, so we don't loose it
							insert(heap, element.score, element.column);
							abort = true;
							#pragma omp flush (abort)
						}

						currentColOfC = element.column;
					}
				}

				if (abort) {
					continue;
				}

				uint64_t score = 0;
				uint64_t heapScore = 0;
				uint64_t *restrict x = checkedCallocUInt64_t(X->integerPerRow);

				// iterate over all entries of A and check how many 1s can be covered
				// with the currentCol-th column of C
				for (uint64_t colOfA=0; colOfA < A->nCols; ++colOfA) {
					uint64_t heapScoreCol = 0;
					uint64_t overcoveredScore = 0;

					for (uint64_t i=0; i < A->integerPerCol; ++i) {
						// find the newly covered 1s
						uint64_t newlyCovered = ~(Covered->matrixCol[colOfA*Covered->integerPerCol + i]);
						newlyCovered &= A->matrixCol[colOfA*A->integerPerCol + i];
						newlyCovered &= C->matrixCol[currentColOfC*C->integerPerCol + i];

						// find the newly overcovered 1s, i.e. 0-entries of A which were not
						// overcovered before
						uint64_t newlyOvercovered = A->matrixCol[colOfA * Covered->integerPerCol + i];
						newlyOvercovered |= Covered->matrixCol[colOfA * Covered->integerPerCol + i];
						newlyOvercovered = ~newlyOvercovered;
						newlyOvercovered &= C->matrixCol[currentColOfC * C->integerPerCol + i];

						heapScoreCol += POPCNT64(newlyCovered) * weight;
						overcoveredScore += POPCNT64(newlyOvercovered);
					}

					heapScore += heapScoreCol;

					long scoreCol = heapScoreCol;
					scoreCol -= overcoveredScore;
					if (scoreCol > 0) {
						score += scoreCol;
						x[colOfA / 64] |= (ONE << colOfA % 64);
					}
				}

				#pragma omp critical(dataupdate)
				{
					insert(updatedHeap, heapScore, currentColOfC);

					// the second condition is to determine the behaviour in case of
					// multiple best columns
					if (score > bestScore || (score > 0 && score == bestScore && currentColOfC < bestCol)) {
						bestScore = score;
						bestCol = currentColOfC;
						memcpy(bestX, x, sizeof(uint64_t) * X->integerPerRow);
					}
				}

				free(x);
			}
		}
		
		if (bestCol == UINT64_MAX) {
			// no column was of any help anymore
			break;
		} else {
//			printf("picking column %d for k=%d\n", (int)bestCol, (int)l);

			// B(:,l) = C(:,bestCol);
			memcpy(B->matrixCol + l*B->integerPerCol,
				   C->matrixCol + bestCol*C->integerPerCol,
				   sizeof(uint64_t)*C->integerPerCol);

			// set the l-th column of x
			memcpy(X->matrixRow + l*X->integerPerRow,
				   bestX,
				   sizeof(uint64_t) * X->integerPerRow);

			// update Covered
			for (uint64_t j=0; j<A->nCols; ++j) {
				// if the i-th entry of bestX is 0,
				// we do not cover anything new in this column
				if ( (bestX[j/64] & (ONE << j%64)) == 0) {
					continue;
				}
				
				for (uint64_t i=0; i<A->integerPerCol; ++i) {
					Covered->matrixCol[j*Covered->integerPerCol + i] |= B->matrixCol[l*B->integerPerCol + i];
				}
			}

			// merge both heaps
			deleteColumn(updatedHeap, bestCol); // delete max element first, since we just used it
			while (updatedHeap->numberOfElements != 0) {
				HeapElement element = deleteMax(updatedHeap);
				insert(heap, element.score, element.column);
			}
		}
	}

	free(bestX);
	freeBitMatColWise(Covered);

	freeBinaryHeap(heap);
	freeBinaryHeap(updatedHeap);
}

