#include "config.h"
#include "cmdline.h"
#include "basso_strto.h"
#include "basso.h"
#include "BasicMatrixOperations.h" // These should be removed
#include "MemoryManagement.h"      // when they're included in basso.h
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#ifdef HAVE_CLOCK_GETTIME
#  include <time.h>
#else
#  ifdef __MACH__
#    include <sys/time.h>
#    define CLOCK_REALTIME 0 
#    define CLOCK_MONOTONIC 0
#  endif /* __MACH__ */
#endif /* HAVE_CLOCK_GETTIME */
#ifdef _OPENMP /*We might want to avoid OpenMP */
#  include <omp.h>
#else
#  define omp_set_num_threads(a) fprintf(stderr, "Warning: Basso was compiled on a system with no OpenMP support. Setting the number of threads has no effect.\n")
#endif /* _OPENMP */







/*
 * Global variables to store the command-line arguments.
 * All parameters set by command-line arguments are here for consistency.
 */

static char *outfile         = NULL; // filename template for output data (-o)
static char informat         = 0;    // the format is either 'm' for MatrixMarket of 'b' for binary
static char outformat        = 0;    // or 0 for unknown (-f and -F)
static size_t n_infiles      = 0;    // number of input files
static char **infile_list    = NULL; // A list of input file names     

static char *infile          = NULL; // filename for current input data, set in main()

static uint64_t rank         = 0;    // The rank of the decomposition (-k)
static double threshold      = -1.0; // The rounding threshold (-t)
static uint64_t weight       = 1;    // The bias weight (-w)
static int processes         = -1;   // Number of threads (-j);
                                     // non-positive numbers mean default behaviour of OpenMP

static int version_flag      = 0;    // should we print the version string
static int verbosity_flag    = 0;    // should we print output (-v)
static int clock_flag        = 0;    // should timing information be printed (-c)
static char *clock_outfile   = NULL; // file for storing the timing iformation


/* Global variable for checking if we've received SIGTERM or SIGINT */
static volatile sig_atomic_t terminate_signal_received = 0;


/*
 * Static functions
 */

static void
print_help(const char *programName)
{
  fprintf(stdout,
          "Decompose Boolean matrix INPUT using the Asso algorithm.\n"
          "\n"
          "Usage: %s -k RANK -t THR [OPTIONS] [-o OUTPUT] [INPUT]\n"
          "\n"
          "Options:\n"
          "  -k, --rank=RANK      set the rank of the factorization\n"
          "  -t, --threshold=TRH  set the rounding threshold from [0, 1]\n"
          "  -w, --weight=WEIGHT  set the bias weight\n"
          "  -f, --format=FMT     set the file format; \"m\" means\n"
          "                       MatrixMarket and \"b\" binary format.\n"
          "                       This is used for input and output\n"
          "                       unless \"-F\" is also given.\n"
          "  -F, --outformat=FMT  set the output file format to FMT\n"
          "  -j, --processes[=N]  set the number of parallel processes;\n"
          "                       if no parameter is given, forces using\n"
          "                       OpenMP's default behaviour\n"
          "  -c, --clock[=FILE]   print timing information to FILE, if given,\n"
          "                       or to standard output\n"
          "\n"
          "  -o, --output=FILE    set the output filename template\n"
          "\n"
          "      --verbose        print information to stderr\n"
          "      --quiet          suppress printing of information (default)\n"
          "\n"
          "  -h, --help           show this help and exit\n"
          "      --version        show the version information and exit\n"
          "\n"
          "\n"
          "Input:"
          "  The input file must in either sparse MatrixMarket format or\n"
          "  in binary format. If \"-f\" option is not given, the format\n"
          "  is guessed from the file extension such that \".mtx\"\n"
          "  denotes MatrixMarket files and \".bbm\" denotes binary files.\n"
          "  If no INPUT is given, or it is \"-\", the input is read from\n"
          "  standard input. In that case, \"-f\" option must be given.\n"
          "Output:\n"
          "  The \"-o\" option sets the output template. The final output files\n"
          "  are called <file>_L.<ext> and <file>_R.<ext> for left and right\n"
          "  factor matrix, respectively. Here, <file> is the argument for \"-o\",\n"
          "  and <ext> is either \"bbm\" or \"mtx\", depending on the output\n"
          "  format. The output format is the same as the input format unless\n"
          "  the \"-F\" option is used to override it. If \"-o\" is not given or\n"
          "  is \"-\", both factor matrices are written to the standard output.\n"
          "Timing:\n"
          "  If the \"-c\" option is given, the program prints timing results.\n"
          "  If no file is given in the argument, standard output is used.\n"
          "  Output is a row of 8 TAB separated numbers giving the start and\n"
          "  stop times in seconds starting from 0 for reading the input file,\n"
          "  computing the association matrix, building the factorization, and\n"
          "  writing the output. The times are wall-clock times.\n"
          "  N.B. The ouput is always appended to the file, not overwritten.\n"
          "\n"
          "Examples:\n"
          "  %s -k10 -t0.8 -o factor matrix.mtx\n"
          "                        decomposes MatrixMarket matrix to\n"
          "                        rank-10 decomposition and prints the\n"
          "                        output to factor_L.mtx and factor_R.mtx\n"
          "  %s -k10 -t0.8 -w2 -j10 -fm -Fb -o factor matrix.txt\n"
          "                        like above, but decomposition values\n"
          "                        covering 1s twice as much as not covering\n"
          "                        0s, 10 parallel threads are used, and the\n"
          "                        input is in MatrixMarket format, but the\n"
          "                        output is written in binary format\n"
          "  %s -k10 -t0.8 -j10 -c times.txt matrix.mtx > /dev/null\n"
          "                        Stores the timing information to times.txt\n"
          "                        and discards the output.\n"
          ,
          programName, // For usage
          programName, // For example 1
          programName, // For example 2
          programName  // For example 3
          );
}

static void
print_version()
{
    fprintf(stdout,
            "%s command line interface %s\n"
            "Copyright (C) 2015 Stefan Neumann & Pauli Miettinen\n"
            "License Apache: The Apache Software Foundation license <http://www.apache.org/licenses>\n"
            "This is free software: you are free to change and redistribute it.\n"
            "There is no warranty.\n"
	    ,
            PACKAGE_NAME, PACKAGE_VERSION);
}



/*
 * Parses the command-line arguments and stores the results to global variables
 */

static int
parse_commandline_arguments(int argc, char *argv[])
{
  int c;

  /*
   * Parse command-line arguments
   */

  while (1) {
    // Arguments
    static struct option long_options[] =
    {
      {"output", required_argument, 0, 'o'},
      {"format", required_argument, 0, 'f'},
      {"outformat", required_argument, 0, 'F'},
      {"rank", required_argument, 0, 'k'},
      {"threshold", required_argument, 0, 't'},
      {"weight", required_argument, 0, 'w'},
      {"processes", optional_argument, 0, 'j'},
      {"clock", optional_argument, 0, 'c'},
      {"verbose", no_argument, &verbosity_flag, 1},
      {"quiet", no_argument, &verbosity_flag, 0},
      {"help", no_argument, 0, 'h'},
      {"version", no_argument, &version_flag, 1},
      {0, 0, 0, 0}
    };
    const char optionstring[] = "o:f:F:k:t:w:j::c::h";
    int option_index = 0;

    c = getopt_long(argc, argv, optionstring, long_options, &option_index);

    if (c == -1) {
      // The end of options
      break;
    }

    switch (c) {
    case 0:
        /* Check if this is version option */
        if (!strcmp(long_options[option_index].name, "version")) {
          print_version();
          exit(BASSO_EXIT_SUCCESS);
        }
        /* No need to do anything for the verbosity flag */
        break;
    case 'h':
      print_help(argv[0]);
      exit(BASSO_EXIT_SUCCESS);

    case 'o':
      outfile = optarg;
      break;

    case 'f':
      informat = optarg[0];
      break;

    case 'F':
      outformat = optarg[0];
      break;

    case 'k':
      rank = strtoullOrExit(optarg, "rank");
      break;

    case 't':
      threshold = strtodOrExit(optarg, "threshold");
      break;

    case 'w':
      weight = strtoullOrExit(optarg, "weight");
      break;

    case 'j':
      if (optarg != NULL) processes = strtolOrExit(optarg, "processes");
      else processes = -1;
      break;

    case 'c':
      clock_flag = 1;
      if (optarg != NULL) clock_outfile = optarg;
      break;
        
    case '?':
      
    default:
      //print_help(argv[0]);
      exit(BASSO_EXIT_INVALID_PARAMETER);
    }
  }

  if (optind == argc) {
    /* Set the infile_list to contain one element, "-" */
    char *str = malloc(2);
    if (str == NULL) {
      fprintf(stderr, "Error when allocating memory for \"-\": %s", strerror(errno));
      exit(BASSO_EXIT_MEMORY);
    }
    strcpy(str, "-");
    infile_list = &str;
    n_infiles = 1;
  } else if (optind < argc) {
    n_infiles = argc - optind;
    infile_list = argv + optind;
  }

  /* Check that rank and threshold are set */
  if (rank == 0) {
    fprintf(stderr, "Error: The rank of the decomposition is not set or is 0.\n"
	    "Use %s -h to get help.\n", argv[0]);
    exit(BASSO_EXIT_INVALID_PARAMETER);
  }

  if (threshold < 0 || threshold > 1) {
    fprintf(stderr, "Error: The rounding threshold is not set or is invalid.\n"
	    "The rounding threshold must be between 0 and 1.\n"
	    "Use %s -h to get help.\n", argv[0]);
    exit(BASSO_EXIT_INVALID_PARAMETER);
  }

  return BASSO_RETURN_SUCCESS;
}

/*
 * Replacement of clock_gettime for OS X
 * This is really computed in microseconds, not in nanoseconds
 */
#ifndef HAVE_CLOCK_GETTIME
int
clock_gettime(int clk_id, struct timespec* t)
{
  struct timeval now;
  int rv = gettimeofday(&now, NULL);
  if (rv) return rv;
  t->tv_sec = now.tv_sec;
  t->tv_nsec = now.tv_usec * 1000;
  return 0;
}
#endif /* undefined HAVE_CLOCK_GETTIME */

/*
 * A standard function that's not part of the C99 standard...
 * The output must be free'd.
 */
static char *
basso_strdup(const char *restrict str)
{
  size_t n = strlen(str) + 1;
  char *dup = malloc(n);
  if (dup == NULL) {
    fprintf(stderr, "Error when allocating memory to duplicate string: %s", strerror(errno));
    exit(BASSO_EXIT_MEMORY);
  }
  memcpy(dup, str, n);
  return dup;
}

/*
 * Signal handler for SIGTERM and SIGINT that kindly asks selectBasis to stop.
 */
void
stop_select_basis(int signum)
{
  basso_stop_select_basis = 1; // This is defined in SelectBasis.h
  terminate_signal_received = 1;
}

/*
 * Returns the string after the last dot in <filename>; returns NULL if there are no dots
 * The output must be free'd.
 */
static char *
get_file_suffix(const char *restrict filename)
{
  char *str, *token, *newtoken;
  const char delimiters[] = ".";
  
  /* We need a writable copy */
  str = basso_strdup(filename);
  /* First, remove anything before the first delimiter (if any) */
  token = strtok(str, delimiters);
  newtoken = strtok(NULL, delimiters);
  do {
    token = newtoken;
  } while ((newtoken = strtok(NULL, delimiters)) != NULL); 

  if (token != NULL) {
    /* We need to allocate new memory that we return to make sure it can be freed */
    char *out = basso_strdup(token);
    token = out;
  }
  free(str);

  return token;
}

/*
 * Returns <filename> with everything after and including the last . (dot) removed;
 * returns <filename> if there are no dots.
 * The output must be free'd.
 */

static char *
strip_file_suffix(const char *restrict filename)
{
  char *suffix = get_file_suffix(filename);
  char *out = basso_strdup(filename);

  if (suffix != NULL) {
    size_t n = strlen(suffix);
    size_t m = strlen(filename);
    out[m-n-1] = '\0'; // -1 to overwrite the dot
  }

  free(suffix);

  return out;
}

/*
 * Builds an output file name of style
 * <prefix><midfix><side><suffix>
 * where <side> is either "_L" or "_R".
 *
 * If <prefix> is NULL, returns "-"; if <midfix> is NULL, it's omitted
 *
 * The output must be free'd.
 */
static char *
get_outfilename(const char *restrict prefix, 
		const char *restrict midfix, 
		const enum factor_side fside, 
		const char *restrict suffix)
{
  char *out;
  char *side;
  size_t prefix_len, midfix_len, side_len, suffix_len, out_len;

  if (prefix == NULL) {
    /* We must allocate this so it can be free'd */
    out = malloc(2);
    if (out == NULL) {
      fprintf(stderr, "Error when allocating memory for output file name: %s", strerror(errno));
      exit(BASSO_EXIT_MEMORY);
    }
    strcpy(out, "-");
    return out;
  }

  if (fside == RIGHT) side = "_R";
  else side = "_L";

  prefix_len = strlen(prefix);
  if (midfix != NULL) midfix_len = strlen(midfix);
  else midfix_len = 0;
  side_len = strlen(side);
  suffix_len = strlen(suffix);
  out_len = prefix_len + midfix_len + side_len + 1 + suffix_len + 1; /* space for '.' and '\0' */
  out = malloc(out_len); 
  if (out == NULL) {
    fprintf(stderr, "Error when allocating memory for output file name: %s", strerror(errno));
    exit(BASSO_EXIT_MEMORY);
  }

  strcpy(  out,                                          prefix);
  if (midfix != NULL) {
    strcpy(out + prefix_len,                             midfix);
  }
  strcpy(  out + prefix_len + midfix_len,                side);
  strcpy(  out + prefix_len + midfix_len + side_len,     ".");
  strcpy(  out + prefix_len + midfix_len + side_len + 1, suffix);

  return out;
}

/*
 * Returns a string to be included into the MatrixMarket format header.
 * The output must be free'd
 */
static char *
get_comments_string(const enum factor_side fside)
{
  char *out;
  char *side;
  time_t t;
  struct tm *tm;
  char date[128];
  const size_t out_len=1024; // Maximum lenght of a comment string
  int nchars;

  t = time(NULL);
  tm = localtime(&t);

  if(strftime(date, sizeof(date), "%Y-%m-%d %H:%M:%S %z", tm) == 0) {
    fprintf(stderr, "Error when formatting time for comments\n");
    return NULL;
  }

  if (fside == RIGHT) side = "Right";
  else side = "Left";


  out = malloc(out_len);
  if (out == NULL) {
    fprintf(stderr, "Error when allocating memory for comments: %s", strerror(errno));
    return NULL;
  }

  nchars = snprintf(out, out_len,
                    "%% %s factor matrix returned by basso on %s\n"
                    "%% Parameters:\n"
                    "%%   Input:     %s\n"
                    "%%   Rank:      %llu\n"
                    "%%   Threshold: %f\n"
                    "%%   Weight:    %llu\n"
                    "%%   Processes: %d"
                    , side,  date, infile, rank, threshold, weight, processes);
  if (nchars >= out_len) {
    free(out);
    return NULL;
  }

  return out;
}

static int
print_outfiles(const basso_bmatrix_cm *restrict B, const basso_bmatrix_rm *restrict C)
{
  char *suffix, *midfix, *outfilename, *comments;
  int out;

  /* Outfile names */

  switch(outformat) {
  case 'm':
    suffix = BASSO_MM_SUFFIX;
    break;
  case 'b':
    fprintf(stderr, "Writing of binary files has not yet been implemented\n");
    exit(BASSO_EXIT_UNIMPLEMENTED);
  default:
    if (outformat) fprintf(stderr, "Unknown output file format %c\n", informat);
    else fprintf(stderr, "Output file format not given and could not be guessed\n");
    exit(BASSO_EXIT_INVALID_PARAMETER);
  }

  if (n_infiles == 1) {
    midfix = NULL;
  } else {
    midfix = strip_file_suffix(infile);
  }

  /* Print left file */
  outfilename = get_outfilename(outfile, midfix, LEFT, suffix);
  comments = get_comments_string(LEFT);

  if (verbosity_flag) fprintf(stderr, "  Writing to %s... ", outfilename);

  out = write_mm_matrix_cm(outfilename, B, comments);
  if (out != BASSO_RETURN_SUCCESS) goto cleanup;

  free(outfilename);
  free(comments);

  /* Print right file */
  outfilename = get_outfilename(outfile, midfix, RIGHT, suffix);
  comments = get_comments_string(RIGHT);

  if (verbosity_flag) fprintf(stderr, "and to %s... ", outfilename);

  out = write_mm_matrix_rm(outfilename, C, comments);
  if (out != BASSO_RETURN_SUCCESS) goto cleanup;

  if (verbosity_flag) fprintf(stderr, "done\n");

 cleanup:
  free(midfix);
  free(outfilename);
  free(comments);
  
  switch(out) {
  case BASSO_RETURN_IOFAIL:
    exit(BASSO_EXIT_IOERROR);
  case BASSO_RETURN_IOPERM:
    exit(BASSO_EXIT_INVALID_OUTPUT);
  case BASSO_RETURN_INVALID_INPUT:
    exit(BASSO_EXIT_INVALID_INPUT);
  case BASSO_RETURN_SUCCESS:
    return BASSO_RETURN_SUCCESS;
  default:
    exit(BASSO_EXIT_UNKNOWN_ERROR);
  } 

  // We should never reach this one
  return BASSO_RETURN_SUCCESS;
}

static int
print_timing_results(const struct timespec *timings)
{
  FILE *fp;
  int i, retval;
  
  if (clock_outfile == NULL || !strcmp(clock_outfile, "-")) {
    fp = stdout;
  } else {
    errno = 0;
    fp = fopen(clock_outfile, "a");
    if (fp == NULL) {
      fprintf(stderr, "Error when opening file %s for writing timing output: %s\n",
              clock_outfile, strerror(errno));
      retval = BASSO_RETURN_IOPERM;
      goto cleanup;
    }
  }
  
  for (i=0; i < 8; i++) {
    double diff;
    int r;
    diff = timings[i].tv_sec - timings[0].tv_sec;
    diff += (timings[i].tv_nsec - timings[0].tv_nsec) / 1000000000.0;
    errno = 0;
    r = fprintf(fp, "%.15e%c", diff, (i < 7)?'\t':'\n');
    if (r < 0) {
      fprintf(stderr, "Error when printing timing output to %s: %s\n",
              clock_outfile, strerror(errno));
      retval = BASSO_RETURN_IOFAIL;
      goto cleanup;
    }
  }
  retval = BASSO_RETURN_SUCCESS;
  
cleanup:
  if (fp != NULL && fp != stdout) fclose(fp);
  return retval;
}


int
main(int argc, char *argv[])
{
  size_t i;

  basso_bmatrix_rm *Mr; /* input data in row major */
  basso_bmatrix_cm *Mc; /* input data in col major */
  basso_bmatrix_rm *Ar; /* association matrix in row major */
  basso_bmatrix_cm *Ac; /* association matrix in col major */
  basso_bmatrix_cm *B;  /* left factor matrix */
  basso_bmatrix_rm *C;  /* right factor matrix */

  struct timespec timings[8];

  struct sigaction term_action;
  struct sigaction default_action;

  /* If there's no command-line arguments, we just print version and exit */
  if (argc == 1) {
    print_version();
    return BASSO_EXIT_SUCCESS;
  }

  /* Parse command-line arguments to populate global variables */
  parse_commandline_arguments(argc, argv);

  /* Set the number of threads if the user has specifically asked for it */
  if (processes > 0) {
    omp_set_num_threads(processes);
  }


  /* For-loop over the input files */
  for (i = 0; i < n_infiles && !terminate_signal_received; i++) {
    infile = infile_list[i];

    /* TODO move format inference to separate function and re-infer per file if necessary */

    if (!informat) {
      char *suffix;
      if (outformat) {
	/* Invalid arguments */
	fprintf(stderr, "Error: Output format cannot be set if input format is set\n");
	return BASSO_EXIT_INVALID_PARAMETER;
      }
      if (!strcmp(infile, "-")) {
	fprintf(stderr, "Error: Input format must be given when reading from stdin\n");
	return BASSO_EXIT_INVALID_PARAMETER;
      }
      suffix = get_file_suffix(infile);
      if (suffix != NULL) {
	if (is_suffix_matrixmarket(suffix)) informat = 'm';
	else if (is_suffix_bassobinary(suffix)) informat = 'b';
      }
      free(suffix);
    }
      
    if (informat && !outformat) outformat = informat;

    /* Notice that it is still possible that informat is unset */

    if (verbosity_flag) fprintf(stderr, "BASSO: decomposing %s with %d processes and "
				"rank = %llu, threshold = %f, and weight = %llu\n",
				infile, processes, rank, threshold, weight);
    // DEBUG
    /*
      printf("Input: %s\n", (infile != NULL)?infile:"NULL");
      printf("Output: %s\n", (outfile != NULL)?outfile:"NULL");
      printf("Format: %c\n", (informat != 0)?informat:' ');
      printf("Outformat: %c\n", (outformat != 0)?outformat:' ');
      printf("Rank: %llu\n", rank);
      printf("Weight: %llu\n", weight);
      printf("Processes: %d\n", processes);
    */

    if (verbosity_flag) fprintf(stderr, "  Reading input from file %s\n", infile);
    clock_gettime(CLOCK_REALTIME, timings); // Time when it all started
    errno = 0;
    switch (informat) {
    case 'm':
      Mr = read_mm_matrix(infile);
      break;
    case 'b':
      fprintf(stderr, "Reading of binary files has not yet been implemented\n");
      return BASSO_EXIT_UNIMPLEMENTED;
    default:
      if (informat) fprintf(stderr, "Unknown input file format %c\n", informat);
      else fprintf(stderr, "Input file format not given and could not be guessed\n");
      return BASSO_EXIT_INVALID_PARAMETER;
    }
 
    if (Mr == NULL) return BASSO_EXIT_IOERROR;
 
    clock_gettime(CLOCK_REALTIME, timings+1);
  

    if (verbosity_flag) fprintf(stderr, "  Building association matrix...\t");
    /* Compute association matrix */
    clock_gettime(CLOCK_REALTIME, timings+2); // Time when computing associatin matr started
    Ar = basso_associationMatrix(Mr, threshold);
    clock_gettime(CLOCK_REALTIME, timings+3); // This is time for ending association_matrix
    if (Ar == NULL) return BASSO_EXIT_UNKNOWN_ERROR;

    if (verbosity_flag) fprintf(stderr, "done!\n");

    /* Transpose Ar and Mr and free the unused ones */
    Ac = colWiseMatrixFromRowWiseMatrix(Ar);
    if (Ac == NULL) return BASSO_EXIT_MEMORY;
    freeBitMatRowWise(Ar);
    Mc = colWiseMatrixFromRowWiseMatrix(Mr);
    if (Mc == NULL) return BASSO_EXIT_MEMORY;
    freeBitMatRowWise(Mr);

    /* Set signal handler to allow aborting gracefully from SelectBasis */
    /* The handler stop_select_basis handles both SIGTERM and SIGINT, 
     * so block both and set flags to SA_RESTART
     */
    term_action.sa_handler = stop_select_basis;
    sigemptyset(&(term_action.sa_mask));
    sigaddset(&(term_action.sa_mask), SIGTERM);
    sigaddset(&(term_action.sa_mask), SIGINT);
    term_action.sa_flags = SA_RESTART;
    
    sigaction(SIGTERM, &term_action, &default_action);
    sigaction(SIGINT,  &term_action, &default_action);
    
    /* Select basis */
    if (verbosity_flag) fprintf(stderr, "  Finding basis vectors...\t");
    clock_gettime(CLOCK_REALTIME, timings+4); // This is time for starting selectBasis
    basso_selectBasis(Mc, Ac, rank, weight, &B, &C);
    clock_gettime(CLOCK_REALTIME, timings+5); // This is time for ending selectBasis

    if (B == NULL || C == NULL) return BASSO_EXIT_MEMORY;
    freeBitMatColWise(Mc);
    freeBitMatColWise(Ac);

    if (verbosity_flag) fprintf(stderr, "done!\n");

    /* Print output */
    clock_gettime(CLOCK_REALTIME, timings+6); // Time when writing output starts
    print_outfiles(B, C);
    clock_gettime(CLOCK_REALTIME, timings+7); // Time when writing output ends

    freeBitMatColWise(B);
    freeBitMatRowWise(C);

    /* Print timing results -- This might fail, but we don't care*/
    if (clock_flag) print_timing_results(timings);
    
    /* Revert the signal handler back to default for the begin of the next iteration */
    sigaction(SIGTERM, &default_action, &term_action);
    sigaction(SIGINT,  &default_action, &term_action);

  } /* END FOR over the input files */

  return BASSO_EXIT_SUCCESS;
}
