#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "matrix.h" // from Matlab

#include "BitConversionMatrix.h"
#include "MatlabMemoryManagement.h"
#include "../basso/MemoryManagement.h"
#include "../basso/basso_commons.h"

/**
 * Returns a bitwiseMatrix containing the information from matlabMat.
 * 
 * Remember that the given Matlab matrix is saved in column major
 * format.
 */
basso_bmatrix_rm*
denseMatlabMatToBitMat(const mxLogical *restrict matlabMat,
					   const uint64_t mRows,
					   const uint64_t nCols)
{
	basso_bmatrix_rm* bitMat = checkedCallocBitMatRowWise(mRows, nCols);
	const uint64_t integerPerRow = bitMat->integerPerRow;

	for (uint64_t i=0; i<mRows*nCols; i++) {
		if (matlabMat[i] == 0) {
			continue;
		}

		uint64_t row = i%mRows;
		uint64_t col = i/mRows;

		bitMat->matrixRow[row*integerPerRow + col/64] |= (ONE << col%64);
		bitMat->onesPerRow[row]++;
	}

	return bitMat;
}

/**
 * Returns a matlatMabtrix containing the information from bitMat.
 * The returned matrix is a dense Matlab matrix.
 */
mxLogical*
bitMatToDenseMatlabMat(const basso_bmatrix_rm *restrict bitMat)
{
	const uint64_t mRows = bitMat->mRows;
	const uint64_t nCols = bitMat->nCols;

	mxLogical* matlabMat = (mxLogical *)checkedCallocMatlabMat(mRows*nCols);

	const uint64_t integerPerRow = nCols/64 + (nCols%64 > 0 ? 1 : 0);

	for (uint64_t row=0; row<mRows; row++) {
		for (uint64_t col=0; col<nCols; col++) {
			if ((bitMat->matrixRow[row*integerPerRow + col/64] & (ONE << col%64)) > 0) {
				matlabMat[col*mRows + row] = ONE;
			}
		}
	}

	return matlabMat;
}

/**
 * Takes a sparse matrix stored at matlabMat and then returns
 * a bitwiseMatrix in row-major format.
 *
 * If you wonder how this works and how the sparse matrix is stored,
 * please consult the Matlab documentation:
 *
 * General idea:	http://de.mathworks.com/help/matlab/matlab_external/matlab-data.html#f17318
 * nzmax:			http://de.mathworks.com/help/matlab/apiref/mxgetnzmax.html
 * ir:				http://de.mathworks.com/help/matlab/apiref/mxgetir.html
 * jc:				http://de.mathworks.com/help/matlab/apiref/mxgetjc.html
 */
basso_bmatrix_rm*
sparseMatlabMatToBitMat(const mxArray *restrict matlabMat,
						const uint64_t mRows,
						const uint64_t nCols)
{
	basso_bmatrix_rm* bitMat = checkedCallocBitMatRowWise(mRows, nCols);
	uint64_t integerPerRow = bitMat->integerPerRow;

	mwIndex *restrict ir = mxGetIr(matlabMat);
	mwIndex *restrict jc = mxGetJc(matlabMat);

	for (mwIndex col=0; col<nCols; col++) {
		mwIndex firstNonZeroIndex = jc[col];
		mwIndex lastNonZeroIndex = jc[col+1] - 1;

		if (jc[col] == jc[col+1]) {
			continue;
		}

		for (mwIndex irIndex=firstNonZeroIndex; irIndex <= lastNonZeroIndex && irIndex < jc[nCols]; irIndex++) {
			mwIndex row = ir[irIndex];

			bitMat->matrixRow[row*integerPerRow + col/64] |= (ONE << col%64);
			bitMat->onesPerRow[row]++;
		}
	}

	return bitMat;
}

