#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <stdio.h>
#ifdef _OPENMP
# include <omp.h>
#else
# define omp_set_num_threads(n) do{ \
if((n)>1) mexWarnMsgTxt("Basso was compiled without OpenMP, number of threads is ignored"); \
}while(0)
#endif

#include "matrix.h" // from Matlab
#include "mex.h"

#include "BitConversionMatrix.h"
#include "MatlabMemoryManagement.h"
/* Headers from basso library */
#include "../basso/BasicMatrixOperations.h"
#include "../basso/MemoryManagement.h"
#include "basso.h"

#ifndef HAVE_MXISSCALAR
/* mxIsScalar is available only from 2015a */
static inline int
mxIsScalar(const mxArray *pm)
{
  const mwSize n = mxGetNumberOfDimensions(pm);
  if (n > 2) return 0; // Scalars are always 1-by-1
  const mwSize *dim = mxGetDimensions(pm);
  if (dim[0] != 1 || dim[1] != 1) return 0;
  return 1;
}
#endif

/**
 * Selects a basis to cover a given input matrix A using the columns of
 * A's association matrix C.
 * The input matrix A must be provided in prhs, the output will be stored in plhs.
 *
 * This function is inteded to be called by Matlab.
 */
void
mexFunction(const int nlhs,
			mxArray *plhs[],
			const int nrhs,
			const mxArray *prhs[])
{
	if (!(nrhs == 3 || nrhs == 4)) {
        mexErrMsgTxt("Wrong number of input arguments. Three or four arguments expected: "
						"A - the input matrix, k - the number of basis vectors, tau - the threshold"
						" and n - the (maximal) number of threads to use");
	}
	if (! mxIsLogical(prhs[0])) {
        mexErrMsgTxt("The input matrix must be logical.");	
	}

    if (mxGetNumberOfDimensions(prhs[0]) != 2) {
        mexErrMsgTxt("The first input variable must be a matrix.");
	}
	if (! mxIsDouble(prhs[1]) || ! mxIsScalar(prhs[1])) {
        mexErrMsgTxt("Second argument must be an integer.");
	}
	if (! mxIsDouble(prhs[2]) || ! mxIsScalar(prhs[2])) {
        mexErrMsgTxt("Third argument must be a scalar of type double.");
	}

    if (nlhs != 2) {
        mexErrMsgTxt("Wrong number of output arguments. Exactly two arguments expected.");
	}

	const uint64_t k = (uint64_t)mxGetScalar(prhs[1]);
	const double threshold = mxGetScalar(prhs[2]);
  int nthreads = 1; // The default
  if (nrhs > 3) {
    if (!mxIsScalar(prhs[3]) || !mxIsDouble(prhs[3])) {
      mexErrMsgTxt("Fourth argument must be an integer.");
    }
    nthreads = (int)mxGetScalar(prhs[3]);
  }
  
	
	const mwSize * const dim = mxGetDimensions(prhs[0]);
	const mwSize mRows = dim[0];
	const mwSize nCols = dim[1];

	if (k >= mRows || k >= nCols) {
        mexErrMsgTxt("Please choose k < min{m, n}, where m and n are the dimensions of the input matrix A."
					" For k >= min{m,n} you could just use a trivial factorization.");
	}

	// set up OpenMP
	omp_set_num_threads(nthreads);

	// Convert to bitwise representation
	basso_bmatrix_rm* Arow;
	if (mxIsSparse(prhs[0])) {
		Arow = sparseMatlabMatToBitMat(prhs[0], mRows, nCols);
	} else {
		mxLogical *matlabMat = mxGetLogicals(prhs[0]);
		Arow = denseMatlabMatToBitMat(matlabMat, mRows, nCols);
	}

	basso_bmatrix_cm *Bcol;
	basso_bmatrix_rm *X;
	basso(Arow, k, threshold, 1, &Bcol, &X);

	freeBitMatRowWise(Arow);

	basso_bmatrix_rm *B = rowWiseMatrixFromColWiseMatrix(Bcol);
	freeBitMatColWise(Bcol);

	mxLogical *matlabB = bitMatToDenseMatlabMat(B);
	mxLogical *matlabX = bitMatToDenseMatlabMat(X);

	freeBitMatRowWise(B);
	freeBitMatRowWise(X);

	// generate output of B
	mwSize outDims[2];
	outDims[0] = mRows;
	outDims[1] = k;
	mwSize numberOfOutFields = mRows*k;

	plhs[0] = mxCreateLogicalArray(2, outDims);
	mxLogical *pn = mxGetLogicals(plhs[0]);
	for (mwSize index=0; index<numberOfOutFields; index++) {
		pn[index] = matlabB[index];
	}

	// generate output of X
	outDims[0] = k;
	outDims[1] = nCols;
	numberOfOutFields = k*nCols;

	plhs[1] = mxCreateLogicalArray(2, outDims);
	pn = mxGetLogicals(plhs[1]);
	for (mwSize index=0; index<numberOfOutFields; index++) {
		pn[index] = matlabX[index];
	}
}
